import ipaddress
from datetime import datetime
from django.db import models
from django.urls import reverse, get_resolver, get_urlconf
from django.utils.text import slugify
from django.contrib.auth.models import Group
from django.core.exceptions import ValidationError

from django_inbound.services import get_resolved_urls


class InboundRule(models.Model):
    '''
        Model of Inbound Rule
    '''
    name = models.CharField(max_length=100)
    alias = models.CharField(max_length=100, null=True, blank=True)
    slug = models.CharField(max_length=100, null=True, blank=True)
    namespace = models.CharField(max_length=100, null=True, blank=True)
    url_name = models.CharField(max_length=100, null=True, blank=True)
    group = models.ForeignKey(Group, on_delete=models.CASCADE)
    allow_all = models.BooleanField(default=False, help_text='Allow all User', verbose_name='Allow all IP\'s')
    is_active = models.BooleanField(default=True)
    extra = models.TextField(null=True, blank=True)
    created = models.DateTimeField(default=datetime.now)

    class Meta:
        verbose_name = 'Inbound Rule'
        verbose_name_plural = 'Inbound Rules'

    def __str__(self):
        return self.name

    def clean(self, *args, **kwargs):
        if not self.namespace:
            raise ValidationError({'namespace': "Please provide data"})
        else:
            url_resolver = get_resolver(get_urlconf())
            all_namespace = list(url_resolver.namespace_dict.keys())
            if self.namespace not in all_namespace:
                raise ValidationError({'namespace': "Invalid namesapce"})
        if self.url_name and self.namespace:
            # path = '{0}:{1}'.format(self.namespace, self.url_name)
            # try:
            #     path = reverse(path)
            # except:
            #     raise ValidationError({'namespace': "Invalid namesapce", 'url_name': "Invalid url name"})
            patterns = get_resolver().url_patterns
            corrosponding_urls = get_resolved_urls(url_patterns=patterns, namespace=self.namespace)
            if self.url_name not in corrosponding_urls:
                raise ValidationError({'url_name': "Invalid url name"})
        super(InboundRule, self).clean(*args, **kwargs)

    def save(self, *args, **kwargs):
        self.slug = slugify(self.name)
        super(InboundRule, self).save(*args, **kwargs)


class InboundIP(models.Model):
    '''
        Model of Inbound IP
    '''
    inbound_rule = models.ForeignKey(InboundRule, on_delete=models.CASCADE)
    start_ip = models.GenericIPAddressField()
    end_ip = models.GenericIPAddressField(null=True, blank=True, help_text='For Single or specific IP, Leave it Blank')
    cidr = models.CharField(max_length=20, null=True, blank=True, help_text='CIDR Block', verbose_name='CIDR')
    created = models.DateTimeField(default=datetime.now)

    class Meta:
        verbose_name = 'Inbound IP'
        verbose_name_plural = 'Inbound IPs'

    def __str__(self):
        return self.inbound_rule.name

    def clean(self, *args, **kwargs):
        start_ip = self.start_ip
        if not self.end_ip:
            end_ip = self.start_ip
        else:
            end_ip = self.end_ip
        startip = ipaddress.IPv4Address(start_ip)
        endip = ipaddress.IPv4Address(end_ip)
        try:
            all_blocks = [ipaddr for ipaddr in ipaddress.summarize_address_range(startip, endip)]
            if not all_blocks:
                raise ValidationError('Invalid Range')
        except:
            raise ValidationError({'start_ip': "Invalid range", 'end_ip': "Invalid Range"})
        if len(all_blocks) > 1:
            raise ValidationError({'start_ip': "Provide range for single CIDR", 'end_ip': "Provide range for single CIDR"})
        super(InboundIP, self).clean(*args, **kwargs)

    def save(self, *args, **kwargs):
        if not self.end_ip:
            self.end_ip = self.start_ip
        if self.start_ip and self.end_ip:
            startip = ipaddress.IPv4Address(self.start_ip)
            endip = ipaddress.IPv4Address(self.end_ip)
            all_blocks = [ipaddr for ipaddr in ipaddress.summarize_address_range(startip, endip)]
            if all_blocks:
                all_blocks = all_blocks[0]
                self.cidr = all_blocks.compressed
        super(InboundIP, self).save(*args, **kwargs)



from django.contrib import admin
from django_inbound.models import InboundRule, InboundIP

# Register your models here.


class InboundIPInline(admin.TabularInline):
    ''' Inline of Inbound Rule '''
    model = InboundIP
    readonly_fields = ['cidr']
    extra = 1


class InboundRuleAdmin(admin.ModelAdmin):
    model = InboundRule
    list_display = ['name', 'namespace', 'url_name', 'group', 'allow_all', 'is_active', 'created']
    exclude = ['alias', 'slug']
    list_filter = ['is_active', 'group', 'namespace', 'url_name']
    inlines = [InboundIPInline]


admin.site.register(InboundRule, InboundRuleAdmin)



import ipaddress
from django.urls import resolve
from django.contrib.auth.models import User, Group
from django.db.models import F
from datetime import datetime, timedelta
from collections import OrderedDict

from django_inbound.models import InboundRule


def validate_inbound_rules(request=None, path=None):
    if request:
        namespace = request.resolver_match.namespace
        url_name = request.resolver_match.url_name
    elif path:
        try:
            match = resolve(path)
            namespace = match.namespace
            url_name = match.url_name
        except:
            namespace = None
            url_name = None
    else:
        namespace = None
        url_name = None
    all_inbound_rules = InboundRule.objects.filter(is_active=True)
    combine_rule = all_inbound_rules.filter(namespace=namespace, url_name=url_name)
    if combine_rule:
        return True
    namespace_rule = all_inbound_rules.filter(namespace=namespace, url_name__isnull=True)
    if namespace_rule:
        return True
    return False


def check_user_authentication(request):
    if request.user.is_authenticated:
        try:
            user = User.objects.get(id=request.user.id)
        except:
            user = None
    else:
        user = None
    return user

def get_groups(user):
    return list(user.groups.all().values_list('id', flat=True))

def get_group_inboundrules(group_ids):
    return InboundRule.objects.filter(group_id__in=group_ids).filter(is_active=True)

def validate_allow_all_ips(inboundrules_queryset):
    if inboundrules_queryset.filter(allow_all=True):
        return True
    else:
        return False

# def validate_ip_in_inbound_ips(requested_ip, inboundrules_queryset):
#     '''may and may not work properly'''
#     output = False
#     for inboundrules in inboundrules_queryset:
#         all_inbound_ips = inboundrules.inboundip_set.filter(start_ip__lte=requested_ip, end_ip__gte=requested_ip)
#         if all_inbound_ips:
#             output = True
#             break
#     return output



def validate_ip_in_inbound_ips(requested_ip, inboundrules_queryset):
    output = False
    for inboundrules in inboundrules_queryset:
        all_inbound_ips = inboundrules.inboundip_set.all()
        for inbound_ip in all_inbound_ips:
            if ipaddress.ip_address(requested_ip) in ipaddress.ip_network(inbound_ip.cidr):
                output = True
                break
        if output:
            break
    return output



import os

from django.http import Http404
from django.core.exceptions import ObjectDoesNotExist
from django.shortcuts import redirect
from django_inbound.queries import validate_inbound_rules, check_user_authentication, get_groups, get_group_inboundrules, validate_allow_all_ips, validate_ip_in_inbound_ips



def get_client_ip(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


def restrict_user_middleware(get_response):

    def middleware(request):
        # Check namespace and name in InboundRules
        if not request.user.is_superuser:
            if validate_inbound_rules(path=request.path):
                user = check_user_authentication(request=request)
                if user:
                    group_ids = get_groups(user=user)
                    if group_ids:
                        inboundrules = get_group_inboundrules(group_ids=group_ids)
                        if inboundrules:
                            # passing all inboundrules
                            if validate_allow_all_ips(inboundrules_queryset=inboundrules):
                                # All IPs are allows
                                pass
                            else:
                                # Specific IPs are allows
                                ip = get_client_ip(request=request)
                                if validate_ip_in_inbound_ips(requested_ip=ip, inboundrules_queryset=inboundrules):
                                    pass
                                else:
                                    raise Http404
                        else:
                            # No inbound rules is attached to user-group
                            raise Http404
                    else:
                        raise Http404
                else:
                    raise Http404
            else:
                pass
        else:
            pass

        response = get_response(request)
        return response

    return middleware




def get_resolved_urls(url_patterns, namespace=None):
    url_patterns_resolved = []
    for entry in url_patterns:
        if hasattr(entry, 'url_patterns'):
            if namespace:
                if entry.namespace == namespace:
                    # url_patterns_resolved += get_resolved_urls(
                    #     entry.url_patterns, namespace=namespace)
                    url_patterns_resolved += get_resolved_urls(
                        entry.url_patterns)
            else:
                url_patterns_resolved += get_resolved_urls(
                    entry.url_patterns)

        else:
            if entry.name is not None:
                url_patterns_resolved.append(entry.name)
    return url_patterns_resolved



# Generated by Django 3.0 on 2020-07-11 16:14

import datetime
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0011_update_proxy_permissions'),
    ]

    operations = [
        migrations.CreateModel(
            name='InboundRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
                ('alias', models.CharField(blank=True, max_length=100, null=True)),
                ('slug', models.CharField(blank=True, max_length=100, null=True)),
                ('namespace', models.CharField(blank=True, max_length=100, null=True)),
                ('url_name', models.CharField(blank=True, max_length=100, null=True)),
                ('allow_all', models.BooleanField(default=False, help_text='Allow all User', verbose_name="Allow all IP's")),
                ('is_active', models.BooleanField(default=True)),
                ('extra', models.TextField(blank=True, null=True)),
                ('created', models.DateTimeField(default=datetime.datetime.now)),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='auth.Group')),
            ],
            options={
                'verbose_name': 'Inbound Rule',
                'verbose_name_plural': 'Inbound Rules',
            },
        ),
        migrations.CreateModel(
            name='InboundIP',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start_ip', models.GenericIPAddressField()),
                ('end_ip', models.GenericIPAddressField(blank=True, help_text='For Single or specific IP, Leave it Blank', null=True)),
                ('cidr', models.CharField(blank=True, help_text='CIDR Block', max_length=20, null=True, verbose_name='CIDR')),
                ('created', models.DateTimeField(default=datetime.datetime.now)),
                ('inbound_rule', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_inbound.InboundRule')),
            ],
            options={
                'verbose_name': 'Inbound IP',
                'verbose_name_plural': 'Inbound IPs',
            },
        ),
    ]
